from __future__ import annotations

__all__ = ["figure2html"]

import base64
import io
import math

from matplotlib import pyplot as plt
from matplotlib.axes import Axes


def figure2html(fig: plt.Figure, close_fig: bool = False) -> str:
    r"""Converts a matplotlib figure to a string that can be used in a
    HTML file.

    Args:
        fig (``Figure``): Specifies the figure to convert.
        close_fig (``bool``, optional): If ``True``, the figure is
            closed after it is converted to HTML format.
            Default: ``False``

    Returns:
        str: The converted figure to a string.

    Example usage:

    .. code-block:: pycon

        >>> from matplotlib import pyplot as plt
        >>> from flamme.utils.figure import figure2html
        >>> fig, ax = plt.subplots()
        >>> string = figure2html(fig)
    """
    fig.tight_layout()
    img = io.BytesIO()
    fig.savefig(img, format="png", bbox_inches="tight")
    img.seek(0)
    data = base64.b64encode(img.getvalue()).decode("utf-8")
    if close_fig:
        plt.close(fig)
    return f'<img src="data:image/png;charset=utf-8;base64, {data}">'


def readable_xticklabels(
    ax: Axes,
    max_num_xticks: int = 100,
    xticklabel_max_len: int = 20,
    xticklabel_min: int = 10,
) -> None:
    r"""Updates the tick labels to make them easier to read, in
    particular if the tick labels are dense.

    Args:
        ax (``matplotlib.axes.Axes``):
        max_num_xticks (int, optional): Specifies the maximum number
            of ticks to show in the figure. Default: ``100``
        xticklabel_max_len (int, optional): If a tick label has a
            length greater than this value, the tick labels are
            rotated vertically. Default: ``20``
        xticklabel_min (int, optional): If the number of ticks is
            lower than this number the tick labels are rotated
            vertically. Default: ``10``

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> from matplotlib import pyplot as plt
        >>> from flamme.utils.figure import readable_xticklabels
        >>> fig, ax = plt.subplots()
        >>> ax.hist(np.arange(10), bins=10)
        >>> readable_xticklabels(ax)
    """
    xticks = ax.get_xticks()
    if len(xticks) > max_num_xticks:
        n = math.ceil(len(xticks) / max_num_xticks)
        xticks = xticks[::n]
        ax.set_xticks(xticks)
    if len(xticks) > xticklabel_min or any(
        [len(str(label)) > xticklabel_max_len for label in ax.get_xticklabels()]
    ):
        ax.tick_params(axis="x", labelrotation=90)



from __future__ import annotations

__all__ = ["DataTypeSection"]

import copy
import logging
from collections.abc import Sequence

import numpy as np
from jinja2 import Template

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)

logger = logging.getLogger(__name__)


class DataTypeSection(BaseSection):
    r"""Implements a section that analyzes the data type of each column.

    Args:
        dtypes (``dict``): Specifies the data type for each column.
        types (``dict``): Specifies the types of the values in each
            column. A column can contain multiple types. The keys are
            the column names.
    """

    def __init__(self, dtypes: dict[str, np.dtype], types: dict[str, set]) -> None:
        self._dtypes = dtypes
        self._types = types

        dtkeys = set(self._dtypes.keys())
        tkeys = set(self._types.keys())
        if dtkeys != tkeys:
            raise RuntimeError(
                f"The keys of dtypes and types do not match:\n"
                f"({len(dtkeys)}): {dtkeys}\n"
                f"({len(tkeys)}): {tkeys}\n"
            )

    def get_statistics(self) -> dict:
        return copy.deepcopy(self._types)

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "table": self._create_table(),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        logger.info("Rendering the data types report...")
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the values types for each column.

<ul>
  <li> <b>data type</b>: is the pandas data type used to represent the column </li>
  <li> <b>types</b>: are the real object types for the objects in the column </li>
</ul>

{{table}}
"""

    def _create_table(self) -> str:
        columns = sorted(self._types.keys())
        rows = "\n".join(
            [
                create_table_row(column=col, types=self._types[col], dtype=self._dtypes[col])
                for col in columns
            ]
        )
        return Template(
            """
<table class="table table-hover table-responsive w-auto" >
    <thead class="thead table-group-divider">
        <tr>
            <th>column</th>
            <th>data type</th>
            <th>types</th>
        </tr>
    </thead>
    <tbody class="tbody table-group-divider">
        {{rows}}
        <tr class="table-group-divider"></tr>
    </tbody>
</table>
"""
        ).render({"rows": rows})


def create_table_row(column: str, dtype: np.dtype, types: set) -> str:
    r"""Creates the HTML code of a new table row.

    Args:
        column (str): Specifies the column name.
        dtype (``numpy.ndtype``): Specifies the column data type.
        types (set): Specifies the types in th column.

    Returns:
        str: The HTML code of a row.
    """
    types = sorted([str(t).replace("<", "&lt;").replace(">", "&gt;") for t in types])
    return Template(
        """<tr>
    <th>{{column}}</th>
    <td>{{dtype}}</td>
    <td>{{types}}</td>
</tr>"""
    ).render({"column": column, "dtype": dtype, "types": ", ".join(types)})



from __future__ import annotations

__all__ = ["ColumnTemporalNullValueAnalyzer"]

import logging

from pandas import DataFrame

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import ColumnTemporalNullValueSection, EmptySection

logger = logging.getLogger(__name__)


class ColumnTemporalNullValueAnalyzer(BaseAnalyzer):
    r"""Implements an analyzer to show the temporal distribution of null
    values for a given column.

    Args:
        column (str): Specifies the column to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.
        figsize (``tuple`` , optional): Specifies the figure size in
            inches. The first dimension is the width and the second is
            the height. Default: ``None``

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.analyzer import TemporalNullValueAnalyzer
        >>> analyzer = ColumnTemporalNullValueAnalyzer(
        ...     column="col", dt_column="datetime", period="M"
        ... )
        >>> analyzer
        ColumnTemporalNullValueAnalyzer(column=col, dt_column=datetime, period=M, figsize=None)
        >>> df = pd.DataFrame(
        ...     {
        ...         "col": np.array([np.nan, 1, 0, 1]),
        ...         "datetime": pd.to_datetime(
        ...             ["2020-01-03", "2020-02-03", "2020-03-03", "2020-04-03"]
        ...         ),
        ...     }
        ... )
        >>> section = analyzer.analyze(df)
    """

    def __init__(
        self,
        column: str,
        dt_column: str,
        period: str,
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._column = column
        self._dt_column = dt_column
        self._period = period
        self._figsize = figsize

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(column={self._column}, "
            f"dt_column={self._dt_column}, period={self._period}, figsize={self._figsize})"
        )

    def analyze(self, df: DataFrame) -> ColumnTemporalNullValueSection | EmptySection:
        logger.info(
            f"Analyzing the temporal null value distribution of {self._column} | "
            f"datetime column: {self._dt_column} | period: {self._period}"
        )
        if self._column not in df:
            logger.info(
                "Skipping temporal null value analysis because the column "
                f"({self._column}) is not in the DataFrame: {sorted(df.columns)}"
            )
            return EmptySection()
        if self._dt_column not in df:
            logger.info(
                "Skipping temporal null value analysis because the datetime column "
                f"({self._dt_column}) is not in the DataFrame: {sorted(df.columns)}"
            )
            return EmptySection()
        return ColumnTemporalNullValueSection(
            df=df,
            column=self._column,
            dt_column=self._dt_column,
            period=self._period,
            figsize=self._figsize,
        )



from __future__ import annotations

__all__ = [
    "BaseDataFrameTransformer",
    "Column",
    "ColumnDataFrameTransformer",
    "ColumnSelection",
    "ColumnSelectionDataFrameTransformer",
    "DecimalToNumeric",
    "DecimalToNumericDataFrameTransformer",
    "NullColumn",
    "NullColumnDataFrameTransformer",
    "Sequential",
    "SequentialDataFrameTransformer",
    "StripString",
    "StripStringDataFrameTransformer",
    "ToDatetime",
    "ToDatetimeDataFrameTransformer",
    "ToNumeric",
    "ToNumericDataFrameTransformer",
    "is_dataframe_transformer_config",
    "setup_dataframe_transformer",
]

from flamme.transformer.df.base import (
    BaseDataFrameTransformer,
    is_dataframe_transformer_config,
    setup_dataframe_transformer,
)
from flamme.transformer.df.column import ColumnDataFrameTransformer
from flamme.transformer.df.column import ColumnDataFrameTransformer as Column
from flamme.transformer.df.datetime import ToDatetimeDataFrameTransformer
from flamme.transformer.df.datetime import ToDatetimeDataFrameTransformer as ToDatetime
from flamme.transformer.df.decimal import DecimalToNumericDataFrameTransformer
from flamme.transformer.df.decimal import (
    DecimalToNumericDataFrameTransformer as DecimalToNumeric,
)
from flamme.transformer.df.null import NullColumnDataFrameTransformer
from flamme.transformer.df.null import NullColumnDataFrameTransformer as NullColumn
from flamme.transformer.df.numeric import ToNumericDataFrameTransformer
from flamme.transformer.df.numeric import ToNumericDataFrameTransformer as ToNumeric
from flamme.transformer.df.selection import ColumnSelectionDataFrameTransformer
from flamme.transformer.df.selection import (
    ColumnSelectionDataFrameTransformer as ColumnSelection,
)
from flamme.transformer.df.sequential import SequentialDataFrameTransformer
from flamme.transformer.df.sequential import (
    SequentialDataFrameTransformer as Sequential,
)
from flamme.transformer.df.string import StripStringDataFrameTransformer
from flamme.transformer.df.string import StripStringDataFrameTransformer as StripString



from __future__ import annotations

__all__ = ["ColumnContinuousAnalyzer", "ColumnTemporalContinuousAnalyzer"]

import logging

from pandas import DataFrame

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import (
    ColumnContinuousSection,
    ColumnTemporalContinuousSection,
    EmptySection,
)

logger = logging.getLogger(__name__)


class ColumnContinuousAnalyzer(BaseAnalyzer):
    r"""Implements an analyzer to show the temporal distribution of
    continuous values.

    Args:
        column: Specifies the column name.
        nbins: Specifies the number of bins in the histogram.
        yscale: Specifies the y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'/'symlog'`` scale is chosen based
            on the distribution.
        xmin: Specifies the minimum value of the range or its
            associated quantile. ``q0.1`` means the 10% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        xmax: Specifies the maximum value of the range or its
            associated quantile. ``q0.9`` means the 90% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        figsize: Specifies the figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.analyzer import ColumnContinuousAnalyzer
        >>> analyzer = ColumnContinuousAnalyzer(column="float")
        >>> analyzer
        ColumnContinuousAnalyzer(column=float, nbins=None, yscale=auto, xmin=q0, xmax=q1, figsize=None)
        >>> df = pd.DataFrame(
        ...     {
        ...         "int": np.array([np.nan, 1, 0, 1]),
        ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
        ...         "str": np.array(["A", "B", None, np.nan]),
        ...     }
        ... )
        >>> section = analyzer.analyze(df)
    """

    def __init__(
        self,
        column: str,
        nbins: int | None = None,
        yscale: str = "auto",
        xmin: float | str | None = "q0",
        xmax: float | str | None = "q1",
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._column = column
        self._nbins = nbins
        self._yscale = yscale
        self._xmin = xmin
        self._xmax = xmax
        self._figsize = figsize

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(column={self._column}, nbins={self._nbins}, "
            f"yscale={self._yscale}, xmin={self._xmin}, xmax={self._xmax}, figsize={self._figsize})"
        )

    def analyze(self, df: DataFrame) -> ColumnContinuousSection | EmptySection:
        logger.info(f"Analyzing the continuous distribution of {self._column}")
        if self._column not in df:
            logger.info(
                "Skipping temporal continuous distribution analysis because the column "
                f"({self._column}) is not in the DataFrame: {sorted(df.columns)}"
            )
            return EmptySection()
        return ColumnContinuousSection(
            column=self._column,
            series=df[self._column],
            nbins=self._nbins,
            yscale=self._yscale,
            xmin=self._xmin,
            xmax=self._xmax,
            figsize=self._figsize,
        )


class ColumnTemporalContinuousAnalyzer(BaseAnalyzer):
    r"""Implements an analyzer to show the temporal distribution of
    continuous values.

    Args:
        column: Specifies the column to analyze.
        dt_column: Specifies the datetime column used to analyze
            the temporal distribution.
        period: Specifies the temporal period e.g. monthly or daily.
        yscale: Specifies the y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'/'symlog'`` scale is chosen based
            on the distribution.
        figsize: Specifies the figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> import pandas as pd
    >>> from flamme.analyzer import TemporalNullValueAnalyzer
    >>> analyzer = ColumnTemporalContinuousAnalyzer(
    ...     column="float", dt_column="datetime", period="M"
    ... )
    >>> analyzer
    ColumnTemporalContinuousAnalyzer(column=float, dt_column=datetime, period=M, yscale=auto, figsize=None)
    >>> df = pd.DataFrame(
    ...     {
    ...         "int": np.array([np.nan, 1, 0, 1]),
    ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
    ...         "str": np.array(["A", "B", None, np.nan]),
    ...         "datetime": pd.to_datetime(
    ...             ["2020-01-03", "2020-02-03", "2020-03-03", "2020-04-03"]
    ...         ),
    ...     }
    ... )
    >>> section = analyzer.analyze(df)

    ```
    """

    def __init__(
        self,
        column: str,
        dt_column: str,
        period: str,
        yscale: str = "auto",
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._column = column
        self._dt_column = dt_column
        self._period = period
        self._yscale = yscale
        self._figsize = figsize

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(column={self._column}, "
            f"dt_column={self._dt_column}, period={self._period}, "
            f"yscale={self._yscale}, figsize={self._figsize})"
        )

    def analyze(self, df: DataFrame) -> ColumnTemporalContinuousSection | EmptySection:
        logger.info(
            f"Analyzing the temporal continuous distribution of {self._column} | "
            f"datetime column: {self._dt_column} | period: {self._period}"
        )
        if self._column not in df:
            logger.info(
                "Skipping temporal continuous distribution analysis because the column "
                f"({self._column}) is not in the DataFrame: {sorted(df.columns)}"
            )
            return EmptySection()
        if self._dt_column not in df:
            logger.info(
                "Skipping temporal continuous distribution analysis because the datetime column "
                f"({self._dt_column}) is not in the DataFrame: {sorted(df.columns)}"
            )
            return EmptySection()
        return ColumnTemporalContinuousSection(
            column=self._column,
            df=df,
            dt_column=self._dt_column,
            period=self._period,
            yscale=self._yscale,
            figsize=self._figsize,
        )



from __future__ import annotations

__all__ = ["ColumnDiscreteSection"]

import logging
from collections import Counter
from collections.abc import Sequence

import numpy as np
from jinja2 import Template
from matplotlib import pyplot as plt

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.figure import figure2html, readable_xticklabels

logger = logging.getLogger(__name__)


class ColumnDiscreteSection(BaseSection):
    r"""Implements a section that analyzes a discrete distribution of
    values.

    Args:
        counter: Specifies the counter that represents the discrete
            distribution.
        null_values: Specifies the number of null values.
        column: Specifies the column name.
        max_rows: Specifies the maximum number of rows to show in the
            table.
        yscale: Specifies the y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'`` scale is chosen based on the
            distribution.
        figsize: Specifies the figure size in inches. The first
            dimension is the width and the second is the height.
    """

    def __init__(
        self,
        counter: Counter,
        null_values: int = 0,
        column: str = "N/A",
        max_rows: int = 20,
        yscale: str = "auto",
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._counter = counter
        self._null_values = null_values
        self._column = column
        self._max_rows = int(max_rows)
        self._yscale = yscale
        self._figsize = figsize

        self._total = sum(self._counter.values())

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""The individual figure size in pixels.

        The first dimension is the width and the second is the height.
        """
        return self._figsize

    @property
    def yscale(self) -> str:
        return self._yscale

    def get_statistics(self) -> dict:
        most_common = [(value, count) for value, count in self._counter.most_common() if count > 0]
        return {
            "most_common": most_common,
            "nunique": len(most_common),
            "total": self._total,
        }

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(f"Rendering the discrete distribution of {self._column}")
        stats = self.get_statistics()
        null_values_pct = (
            f"{100 * self._null_values / stats['total']:.2f}" if stats["total"] > 0 else "N/A"
        )
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._column,
                "total_values": f"{stats['total']:,}",
                "unique_values": f"{stats['nunique']:,}",
                "null_values": f"{self._null_values:,}",
                "null_values_pct": null_values_pct,
                "figure": self._create_figure(),
                "table": self._create_table(),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the discrete distribution of values for column <em>{{column}}</em>.

<ul>
  <li> total values: {{total_values}} </li>
  <li> number of unique values: {{unique_values}} </li>
  <li> number of null values: {{null_values}} / {{total_values}} ({{null_values_pct}}%) </li>
</ul>

{{figure}}
{{table}}
<p style="margin-top: 1rem;">
"""

    def _create_figure(self) -> str:
        if self._total == 0:
            return "<span>&#9888;</span> No figure is generated because the column is empty"
        most_common = [(value, count) for value, count in self._counter.most_common() if count > 0]
        fig = create_histogram(
            column=self._column,
            labels=[str(value) for value, _ in most_common],
            counts=[count for _, count in most_common],
            yscale=self._yscale,
            figsize=self._figsize,
        )
        return Template(
            r"""
<p style="margin-top: 1rem;">
<b>Distribution of values in column {{column}}</b>

<p>The values in the figure below are sorted by decreasing order of number of occurrences.

{{figure}}
"""
        ).render({"figure": fig, "column": self._column})

    def _create_table(self) -> str:
        if self._total == 0:
            return "<span>&#9888;</span> No table is generated because the column is empty"

        most_common = self._counter.most_common(self._max_rows)
        rows_head = "\n".join(
            [create_table_row(column=col, count=count) for col, count in most_common]
        )
        lest_common = self._counter.most_common()[-self._max_rows :][::-1]
        rows_tail = "\n".join(
            [create_table_row(column=col, count=count) for col, count in lest_common]
        )
        return Template(
            """
<details>
    <summary>Show head and tail values</summary>

    <div class="row">
      <div class="col">
        <p style="margin-top: 1rem;">
        <b>Head: {{max_values}} most common values in column <em>{{column}}</em></b>
        <table class="table table-hover table-responsive w-auto" >
            <thead class="thead table-group-divider">
                <tr>
                    <th>column</th>
                    <th>count</th>
                </tr>
            </thead>
            <tbody class="tbody table-group-divider">
                {{rows_head}}
                <tr class="table-group-divider"></tr>
            </tbody>
        </table>
      </div>
      <div class="col">
        <p style="margin-top: 1rem;">
        <b>Tail: {{max_values}} least common values in column <em>{{column}}</em></b>
        <table class="table table-hover table-responsive w-auto" >
            <thead class="thead table-group-divider">
                <tr>
                    <th>column</th>
                    <th>count</th>
                </tr>
            </thead>
            <tbody class="tbody table-group-divider">
                {{rows_tail}}
                <tr class="table-group-divider"></tr>
            </tbody>
        </table>
      </div>
    </div>
</details>
"""
        ).render(
            {
                "max_values": len(most_common),
                "rows_head": rows_head,
                "rows_tail": rows_tail,
                "column": self._column,
            }
        )


def create_histogram(
    column: str,
    labels: list[str],
    counts: list[int],
    yscale: str = "auto",
    figsize: tuple[float, float] | None = None,
) -> str:
    fig, ax = plt.subplots(figsize=figsize)
    x = np.arange(len(labels))
    ax.bar(x, counts, width=0.9 if len(labels) < 50 else 1, color="tab:blue")
    if yscale == "auto":
        yscale = "log" if (max(counts) / min(counts)) >= 50 else "linear"
    ax.set_yscale(yscale)
    ax.set_xticks(x, labels=labels)
    readable_xticklabels(ax, max_num_xticks=100)
    ax.set_xlim(-0.5, len(labels) - 0.5)
    ax.set_ylabel("Number of occurrences")
    ax.set_title(f"Number of occurrences for each value of {column}")
    return figure2html(fig, close_fig=True)


def create_table_row(column: str, count: int) -> str:
    r"""Creates the HTML code of a new table row.

    Args:
        column (str): Specifies the column name.
        count (int): Specifies the count for the column.

    Returns:
        str: The HTML code of a row.
    """
    return Template("""<tr><th>{{column}}</th><td {{num_style}}>{{count}}</td></tr>""").render(
        {"num_style": 'style="text-align: right;"', "column": column, "count": f"{count:,}"}
    )



from __future__ import annotations

__all__ = ["TemporalRowCountSection"]

import logging
from collections.abc import Sequence

from jinja2 import Template
from matplotlib import pyplot as plt
from pandas import DataFrame

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.figure import figure2html, readable_xticklabels

logger = logging.getLogger(__name__)


class TemporalRowCountSection(BaseSection):
    r"""Implement a section to analyze the number of rows per temporal
    window.

    Args:
        df: Specifies the DataFrame to analyze.
        dt_column: Specifies the datetime column used to analyze
            the temporal distribution.
        period: Specifies the temporal period e.g. monthly or daily.
        figsize: Specifies the figure size in inches. The first
            dimension is the width and the second is the height.
    """

    def __init__(
        self,
        df: DataFrame,
        dt_column: str,
        period: str,
        figsize: tuple[float, float] | None = None,
    ) -> None:
        if dt_column not in df:
            raise ValueError(
                f"Datetime column {dt_column} is not in the DataFrame "
                f"(columns:{sorted(df.columns)})"
            )
        self._df = df
        self._dt_column = dt_column
        self._period = period
        self._figsize = figsize

    @property
    def df(self) -> DataFrame:
        r"""The DataFrame to analyze."""
        return self._df

    @property
    def dt_column(self) -> str:
        r"""The datetime column."""
        return self._dt_column

    @property
    def period(self) -> str:
        r"""The temporal period used to analyze the data."""
        return self._period

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""The individual figure size in pixels.

        The first dimension is the width and the second is the height.
        """
        return self._figsize

    def get_statistics(self) -> dict:
        return {}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(
            "Rendering the number of rows per temporal window "
            f"| datetime column: {self._dt_column} | period: {self._period}"
        )
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "dt_column": self._dt_column,
                "figure": create_temporal_count_figure(
                    df=self._df,
                    dt_column=self._dt_column,
                    period=self._period,
                    figsize=self._figsize,
                ),
                "table": create_temporal_count_table(
                    df=self._df,
                    dt_column=self._dt_column,
                    period=self._period,
                ),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the number of rows per temporal window.
The column <em>{{dt_column}}</em> is used as the temporal column.

{{figure}}

{{table}}
<p style="margin-top: 1rem;">
"""


def create_temporal_count_figure(
    df: DataFrame,
    dt_column: str,
    period: str,
    figsize: tuple[float, float] | None = None,
) -> str:
    r"""Return a HTML representation of a figure with number of rows per
    temporal windows.

    Args:
        df: Specifies the DataFrame to analyze.
        dt_column: Specifies the datetime column used to analyze
            the temporal distribution.
        period: Specifies the temporal period e.g. monthly or daily.

    Returns:
        str: The HTML representation of the figure.
    """
    if df.shape[0] == 0:
        return "<span>&#9888;</span> No figure is generated because there is no data"

    counts, labels = prepare_data(df=df, dt_column=dt_column, period=period)
    fig, ax = plt.subplots(figsize=figsize)
    ax.bar(x=labels, height=counts, color="tab:blue")
    ax.set_ylabel("number of rows")
    ax.set_xlim(-0.5, len(labels) - 0.5)
    readable_xticklabels(ax, max_num_xticks=100)
    return figure2html(fig, close_fig=True)


def create_temporal_count_table(df: DataFrame, dt_column: str, period: str) -> str:
    r"""Return a HTML representation of a figure with number of rows per
    temporal windows.

    Args:
        df: Specifies the DataFrame to analyze.
        dt_column: Specifies the datetime column used to analyze
            the temporal distribution.
        period: Specifies the temporal period e.g. monthly or daily.

    Returns:
        str: The HTML representation of the table.
    """
    if df.shape[0] == 0:
        return ""
    counts, labels = prepare_data(df=df, dt_column=dt_column, period=period)
    rows = []
    for label, num_rows in zip(labels, counts):
        rows.append(create_temporal_count_table_row(label=label, num_rows=num_rows))
    return Template(
        """
<details>
    <summary>Statistics per period</summary>

    <p>The following table shows some statistics for each period.

    <table class="table table-hover table-responsive w-auto" >
        <thead class="thead table-group-divider">
            <tr>
                <th>period</th>
                <th>number of rows</th>
            </tr>
        </thead>
        <tbody class="tbody table-group-divider">
            {{rows}}
            <tr class="table-group-divider"></tr>
        </tbody>
    </table>
</details>
"""
    ).render({"rows": "\n".join(rows), "period": period})


def create_temporal_count_table_row(label: str, num_rows: int) -> str:
    r"""Return the HTML code of a table row.

    Args:
        label: Specifies the label i.e. temporal window.
        num_rows: Specifies the number of rows for the given temporal
            window.

    Returns:
        The HTML code of a row.
    """
    return Template("<tr><th>{{label}}</th><td {{num_style}}>{{num_rows}}</td></tr>").render(
        {
            "num_style": 'style="text-align: right;"',
            "label": label,
            "num_rows": f"{num_rows:,}",
        }
    )


def prepare_data(
    df: DataFrame,
    dt_column: str,
    period: str,
) -> tuple[list[int], list[str]]:
    r"""Prepare the data to create the figure and table.

    Args:
        df: Specifies the DataFrame to analyze.
        dt_column: Specifies the datetime column used to analyze
            the temporal distribution.
        period: Specifies the temporal period e.g. monthly or daily.

    Returns:
        A tuple with the counts and the temporal window labels.
    """
    if df.shape[0] == 0:
        return [], []

    df = df[[dt_column]].copy()
    columns = df.columns.tolist()
    dt_col = "__datetime__"
    df[dt_col] = df[dt_column].dt.to_period(period)

    df_count = df.groupby(dt_col)[columns].size()
    count = df_count.to_numpy().astype(int).tolist()
    labels = [str(dt) for dt in df_count.index]
    return count, labels



from __future__ import annotations

__all__ = ["find_range"]

import numpy as np


def find_range(
    values: np.ndarray,
    xmin: float | int | str | None = None,
    xmax: float | int | str | None = None,
) -> tuple[float | int, float | int]:
    r"""Finds a range of value.

    Args:
        values (``numpy.ndarray``): Specifies the values used to find
            the quantiles.
        xmin (float or str or None, optional): Specifies the minimum
            value of the range or its associated quantile.
            ``q0.1`` means the 10% quantile. ``0`` is the minimum
            value and ``1`` is the maximum value. Default: ``None``
        xmax (float or str or None, optional): Specifies the maximum
            value of the range or its associated quantile.
            ``q0.9`` means the 90% quantile. ``0`` is the minimum
            value and ``1`` is the maximum value. Default: ``None``

    Returns:
        tuple: The range of values in the format ``(min, max)``.

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> from flamme.utils.range import find_range
        >>> data = np.arange(101)
        >>> find_range(data)
        (0, 100)
        >>> find_range(data, xmin=5, xmax=50)
        (5, 50)
        >>> find_range(data, xmin="q0.1", xmax="q0.9")
        (10.0, 90.0)
    """
    if xmin is None:
        xmin = np.nanmin(values).item()
    if xmax is None:
        xmax = np.nanmax(values).item()
    q = [float(x[1:]) for x in [xmin, xmax] if isinstance(x, str)]
    quantiles = np.nanquantile(values, q)
    if isinstance(xmin, str):
        xmin = quantiles[0]
    if isinstance(xmax, str):
        xmax = quantiles[-1]
    return (xmin, xmax)



from __future__ import annotations

__all__ = ["BaseAnalyzer", "setup_analyzer"]

import logging
from abc import ABC

from objectory import AbstractFactory
from objectory.utils import is_object_config
from pandas import DataFrame

from flamme.section import BaseSection

logger = logging.getLogger(__name__)


class BaseAnalyzer(ABC, metaclass=AbstractFactory):
    r"""Defines the base class to analyze a DataFrame.

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.analyzer import NullValueAnalyzer
        >>> analyzer = NullValueAnalyzer()
        >>> analyzer
        NullValueAnalyzer(figsize=None)
        >>> df = pd.DataFrame(
        ...     {
        ...         "int": np.array([np.nan, 1, 0, 1]),
        ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
        ...         "str": np.array(["A", "B", None, np.nan]),
        ...     }
        ... )
        >>> analyzer.analyze(df)
    """

    def analyze(self, df: DataFrame) -> BaseSection:
        r"""Analyzes the data in a DataFrame.

        Args:
        ----
            df (``pandas.DataFrame``): Specifies the DataFrame with
                the data to analyze.

        Returns:
        -------
            ``BaseSection``: The section report.

        Example usage:

        .. code-block:: pycon

            >>> import numpy as np
            >>> import pandas as pd
            >>> from flamme.analyzer import NullValueAnalyzer
            >>> analyzer = NullValueAnalyzer()
            >>> df = pd.DataFrame(
            ...     {
            ...         "int": np.array([np.nan, 1, 0, 1]),
            ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
            ...         "str": np.array(["A", "B", None, np.nan]),
            ...     }
            ... )
            >>> analyzer.analyze(df)
        """


def is_analyzer_config(config: dict) -> bool:
    r"""Indicates if the input configuration is a configuration for a
    ``BaseAnalyzer``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
        config (dict): Specifies the configuration to check.

    Returns:
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseAnalyzer`` object.

    Example usage:

    .. code-block:: pycon

        >>> from flamme.analyzer import is_analyzer_config
        >>> is_analyzer_config({"_target_": "flamme.analyzer.NullValueAnalyzer"})
        True
    """
    return is_object_config(config, BaseAnalyzer)


def setup_analyzer(
    analyzer: BaseAnalyzer | dict,
) -> BaseAnalyzer:
    r"""Sets up an analyzer.

    The analyzer is instantiated from its configuration
    by using the ``BaseAnalyzer`` factory function.

    Args:
        analyzer (``BaseAnalyzer`` or dict): Specifies an
            analyzer or its configuration.

    Returns:
        ``BaseAnalyzer``: An instantiated analyzer.

    Example usage:

    .. code-block:: pycon

        >>> from flamme.analyzer import setup_analyzer
        >>> analyzer = setup_analyzer({"_target_": "flamme.analyzer.NullValueAnalyzer"})
        >>> analyzer
        NullValueAnalyzer(figsize=None)
    """
    if isinstance(analyzer, dict):
        logger.info("Initializing an analyzer from its configuration... ")
        analyzer = BaseAnalyzer.factory(**analyzer)
    if not isinstance(analyzer, BaseAnalyzer):
        logger.warning(f"analyzer is not a `BaseAnalyzer` (received: {type(analyzer)})")
    return analyzer



from __future__ import annotations

__all__ = ["ColumnDiscreteAnalyzer", "ColumnTemporalDiscreteAnalyzer"]

import logging
from collections import Counter

from pandas import DataFrame

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import (
    ColumnDiscreteSection,
    ColumnTemporalDiscreteSection,
    EmptySection,
)

logger = logging.getLogger(__name__)


class ColumnDiscreteAnalyzer(BaseAnalyzer):
    r"""Implements a discrete distribution analyzer.

    Args:
        column: Specifies the column to analyze.
        dropna: If ``True``, the NaN values are not included in the
            analysis.
        max_rows: Specifies the maximum number of rows to show in the
            table.
        yscale: Specifies the y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'`` scale is chosen based on the
            distribution.
        figsize: Specifies the figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.analyzer import ColumnDiscreteAnalyzer
        >>> analyzer = ColumnDiscreteAnalyzer(column="str")
        >>> analyzer
        ColumnDiscreteAnalyzer(column=str, dropna=False, max_rows=20, yscale=auto, figsize=None)
        >>> df = pd.DataFrame(
        ...     {
        ...         "int": np.array([np.nan, 1, 0, 1]),
        ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
        ...         "str": np.array(["A", "B", None, np.nan]),
        ...     }
        ... )
        >>> section = analyzer.analyze(df)
    """

    def __init__(
        self,
        column: str,
        dropna: bool = False,
        max_rows: int = 20,
        yscale: str = "auto",
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._column = column
        self._dropna = bool(dropna)
        self._max_rows = max_rows
        self._yscale = yscale
        self._figsize = figsize

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(column={self._column}, "
            f"dropna={self._dropna}, max_rows={self._max_rows}, yscale={self._yscale}, "
            f"figsize={self._figsize})"
        )

    def analyze(self, df: DataFrame) -> ColumnDiscreteSection | EmptySection:
        logger.info(f"Analyzing the discrete distribution of {self._column}")
        if self._column not in df:
            logger.info(
                f"Skipping discrete distribution analysis of column {self._column} "
                f"because it is not in the DataFrame: {sorted(df.columns)}"
            )
            return EmptySection()
        return ColumnDiscreteSection(
            counter=Counter(df[self._column].value_counts(dropna=self._dropna).to_dict()),
            null_values=df[self._column].isnull().sum(),
            column=self._column,
            max_rows=self._max_rows,
            yscale=self._yscale,
            figsize=self._figsize,
        )


class ColumnTemporalDiscreteAnalyzer(BaseAnalyzer):
    r"""Implements an analyzer to show the temporal distribution of
    discrete values.

    Args:
        column (str): Specifies the column to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.
        figsize (``tuple`` , optional): Specifies the figure size in
            inches. The first dimension is the width and the second is
            the height. Default: ``None``

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.analyzer import ColumnTemporalDiscreteAnalyzer
        >>> analyzer = ColumnTemporalDiscreteAnalyzer(
        ...     column="str", dt_column="datetime", period="M"
        ... )
        >>> analyzer
        ColumnTemporalDiscreteAnalyzer(column=str, dt_column=datetime, period=M, figsize=None)
        >>> df = pd.DataFrame(
        ...     {
        ...         "int": np.array([np.nan, 1, 0, 1]),
        ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
        ...         "str": np.array(["A", "B", None, np.nan]),
        ...         "datetime": pd.to_datetime(
        ...             ["2020-01-03", "2020-02-03", "2020-03-03", "2020-04-03"]
        ...         ),
        ...     }
        ... )
        >>> section = analyzer.analyze(df)
    """

    def __init__(
        self,
        column: str,
        dt_column: str,
        period: str,
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._column = column
        self._dt_column = dt_column
        self._period = period
        self._figsize = figsize

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(column={self._column}, "
            f"dt_column={self._dt_column}, period={self._period}, figsize={self._figsize})"
        )

    def analyze(self, df: DataFrame) -> ColumnTemporalDiscreteSection | EmptySection:
        logger.info(
            f"Analyzing the temporal discrete distribution of {self._column} | "
            f"datetime column: {self._dt_column} | period: {self._period}"
        )
        if self._column not in df:
            logger.info(
                "Skipping temporal discrete distribution analysis because the column "
                f"({self._column}) is not in the DataFrame: {sorted(df.columns)}"
            )
            return EmptySection()
        if self._dt_column not in df:
            logger.info(
                "Skipping temporal discrete distribution analysis because the datetime column "
                f"({self._dt_column}) is not in the DataFrame: {sorted(df.columns)}"
            )
            return EmptySection()
        return ColumnTemporalDiscreteSection(
            column=self._column,
            df=df,
            dt_column=self._dt_column,
            period=self._period,
            figsize=self._figsize,
        )



from __future__ import annotations

__all__ = ["BaseColumn", "Column"]

from abc import abstractmethod

from coola.utils import str_indent, str_mapping

from flamme.analyzer.base import BaseAnalyzer, setup_analyzer
from flamme.transformer.series.base import (
    BaseSeriesTransformer,
    setup_series_transformer,
)


class BaseColumn:
    r"""Defines the column base class.

    Example usage:

    .. code-block:: pycon

        >>> from flamme.utils.columns import Column
        >>> from flamme.analyzer import ColumnContinuousAnalyzer
        >>> from flamme.transformer.series import ToNumeric
        >>> column = Column(
        ...     can_be_null=True,
        ...     analyzer=ColumnContinuousAnalyzer(column="col"),
        ...     transformer=ToNumeric(),
        ... )
        >>> column
        Column(
          (can_be_null): True
          (analyzer): ColumnContinuousAnalyzer(column=col, nbins=None, yscale=auto, xmin=q0, xmax=q1, figsize=None)
          (transformer): ToNumericSeriesTransformer()
        )
        >>> analyzer = column.get_analyzer()
        >>> analyzer
        ColumnContinuousAnalyzer(column=col, nbins=None, yscale=auto, xmin=q0, xmax=q1, figsize=None)
        >>> transformer = column.get_transformer()
        >>> transformer
        ToNumericSeriesTransformer()
    """

    @abstractmethod
    def get_analyzer(self) -> BaseAnalyzer:
        r"""Gets the column analyzer.

        Returns:
        -------
            ``BaseAnalyzer``: The column analyzer.
        """

    @abstractmethod
    def get_transformer(self) -> BaseSeriesTransformer:
        r"""Gets the column transformer.

        Returns:
        -------
            ``BaseSeriesTransformer``: The column transformer.
        """


class Column(BaseColumn):
    r"""Defines the column base class.

    Args:
        can_be_null (bool): ``True`` if the column can have null
            values, otherwise ``False``.
        analyzer (``BaseAnalyzer`` or dict): Specifies the column
            analyzer or its configuration.
        transformer (``BaseSeriesTransformer`` or dict): Specifies
            the column  transformer or its configuration.

    Example usage:

    .. code-block:: pycon

        >>> from flamme.utils.columns import Column
        >>> from flamme.analyzer import ColumnContinuousAnalyzer
        >>> from flamme.transformer.series import ToNumeric
        >>> column = Column(
        ...     can_be_null=True,
        ...     analyzer=ColumnContinuousAnalyzer(column="col"),
        ...     transformer=ToNumeric(),
        ... )
        >>> column
        Column(
          (can_be_null): True
          (analyzer): ColumnContinuousAnalyzer(column=col, nbins=None, yscale=auto, xmin=q0, xmax=q1, figsize=None)
          (transformer): ToNumericSeriesTransformer()
        )
        >>> analyzer = column.get_analyzer()
        >>> analyzer
        ColumnContinuousAnalyzer(column=col, nbins=None, yscale=auto, xmin=q0, xmax=q1, figsize=None)
        >>> transformer = column.get_transformer()
        >>> transformer
        ToNumericSeriesTransformer()
    """

    def __init__(
        self,
        can_be_null: bool,
        analyzer: BaseAnalyzer | dict,
        transformer: BaseSeriesTransformer | dict,
    ) -> None:
        self._can_be_null = bool(can_be_null)
        self._analyzer = setup_analyzer(analyzer)
        self._transformer = setup_series_transformer(transformer)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "can_be_null": self._can_be_null,
                    "analyzer": self._analyzer,
                    "transformer": self._transformer,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    @property
    def can_be_null(self) -> bool:
        r"""bool: ``True`` if the column can have null values, otherwise ``False``"""
        return self._can_be_null

    def get_analyzer(self) -> BaseAnalyzer:
        return self._analyzer

    def get_transformer(self) -> BaseSeriesTransformer:
        return self._transformer



from __future__ import annotations

__all__ = ["Ingestor"]


from pandas import DataFrame

from flamme.ingestor.base import BaseIngestor


class Ingestor(BaseIngestor):
    r"""Implements a simple DataFrame ingestor.

    Args:
        df (``pandas.DataFrame``): Specifies the DataFrame to ingest.

    Example usage:

    .. code-block:: pycon

        >>> import pandas as pd
        >>> from flamme.ingestor import Ingestor
        >>> ingestor = Ingestor(
        ...     df=pd.DataFrame(
        ...         {
        ...             "col1": [1, 2, 3, 4, 5],
        ...             "col2": ["1", "2", "3", "4", "5"],
        ...             "col3": ["1", "2", "3", "4", "5"],
        ...             "col4": ["a", "b", "c", "d", "e"],
        ...         }
        ...     )
        ... )
        >>> ingestor
        Ingestor(shape=(5, 4))
        >>> df = ingestor.ingest()
    """

    def __init__(self, df: DataFrame) -> None:
        self._df = df

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(shape={self._df.shape})"

    def ingest(self) -> DataFrame:
        return self._df



from __future__ import annotations

__all__ = [
    "BaseDataFrameTransformer",
    "is_dataframe_transformer_config",
    "setup_dataframe_transformer",
]

import logging
from abc import ABC

from objectory import AbstractFactory
from objectory.utils import is_object_config
from pandas import DataFrame

logger = logging.getLogger(__name__)


class BaseDataFrameTransformer(ABC, metaclass=AbstractFactory):
    r"""Defines the base class to transform a ``pandas.DataFrame``.

    Example usage:

    .. code-block:: pycon

        >>> import pandas as pd
        >>> from flamme.transformer.df import ToNumeric
        >>> transformer = ToNumeric(columns=["col1", "col3"])
        >>> transformer
        ToNumericDataFrameTransformer(columns=('col1', 'col3'))
        >>> df = pd.DataFrame(
        ...     {
        ...         "col1": [1, 2, 3, 4, 5],
        ...         "col2": ["1", "2", "3", "4", "5"],
        ...         "col3": ["1", "2", "3", "4", "5"],
        ...         "col4": ["a", "b", "c", "d", "e"],
        ...     }
        ... )
        >>> df.dtypes
        col1     int64
        col2    object
        col3    object
        col4    object
        dtype: object
        >>> df = transformer.transform(df)
        >>> df.dtypes
        col1     int64
        col2    object
        col3     int64
        col4    object
        dtype: object
    """

    def transform(self, df: DataFrame) -> DataFrame:
        r"""Transforms the data in the ``pandas.DataFrame``.

        Args:
        ----
            df (``pandas.DataFrame``): Specifies the
                ``pandas.DataFrame`` to transform.

        Returns:
        -------
            ``pandas.DataFrame``: The transformed DataFrame.

        Example usage:

        .. code-block:: pycon

            >>> import pandas as pd
            >>> from flamme.transformer.df import ToNumeric
            >>> transformer = ToNumeric(columns=["col1", "col3"])
            >>> df = pd.DataFrame(
            ...     {
            ...         "col1": [1, 2, 3, 4, 5],
            ...         "col2": ["1", "2", "3", "4", "5"],
            ...         "col3": ["1", "2", "3", "4", "5"],
            ...         "col4": ["a", "b", "c", "d", "e"],
            ...     }
            ... )
            >>> df = transformer.transform(df)
            >>> df.dtypes
            col1     int64
            col2    object
            col3     int64
            col4    object
            dtype: object
        """


def is_dataframe_transformer_config(config: dict) -> bool:
    r"""Indicates if the input configuration is a configuration for a
    ``BaseDataFrameTransformer``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
        config (dict): Specifies the configuration to check.

    Returns:
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseDataFrameTransformer`` object.

    Example usage:

    .. code-block:: pycon

        >>> from flamme.transformer.df import is_dataframe_transformer_config
        >>> is_dataframe_transformer_config(
        ...     {"_target_": "flamme.transformer.df.ToNumeric", "columns": ["col1", "col3"]}
        ... )
        True
    """
    return is_object_config(config, BaseDataFrameTransformer)


def setup_dataframe_transformer(
    transformer: BaseDataFrameTransformer | dict,
) -> BaseDataFrameTransformer:
    r"""Sets up a ``pandas.DataFrame`` transformer.

    The transformer is instantiated from its configuration
    by using the ``BaseDataFrameTransformer`` factory function.

    Args:
        transformer (``BaseDataFrameTransformer`` or dict): Specifies a
            ``pandas.DataFrame`` transformer or its configuration.

    Returns:
        ``BaseDataFrameTransformer``: An instantiated transformer.

    Example usage:

    .. code-block:: pycon

        >>> from flamme.transformer.df import setup_dataframe_transformer
        >>> transformer = setup_dataframe_transformer(
        ...     {"_target_": "flamme.transformer.df.ToNumeric", "columns": ["col1", "col3"]}
        ... )
        >>> transformer
        ToNumericDataFrameTransformer(columns=('col1', 'col3'))
    """
    if isinstance(transformer, dict):
        logger.info("Initializing a DataFrame transformer from its configuration... ")
        transformer = BaseDataFrameTransformer.factory(**transformer)
    if not isinstance(transformer, BaseDataFrameTransformer):
        logger.warning(
            f"transformer is not a `BaseDataFrameTransformer` (received: {type(transformer)})"
        )
    return transformer



from __future__ import annotations

__all__ = ["TransformedIngestor"]

import logging

from coola.utils import str_indent, str_mapping
from pandas import DataFrame

from flamme.ingestor.base import BaseIngestor, setup_ingestor
from flamme.transformer.df.base import (
    BaseDataFrameTransformer,
    setup_dataframe_transformer,
)

logger = logging.getLogger(__name__)


class TransformedIngestor(BaseIngestor):
    r"""Implements an ingestor that also transforms the DataFrame.

    Args:
        path (``pathlib.Path`` or str): Specifies the path to the
            CSV file to ingest.
        transformer (``BaseDataFrameTransformer`` or dict): Specifies
            a ``pandas.DataFrame`` transformer or its configuration.

    Example usage:

    .. code-block:: pycon

        >>> from flamme.ingestor import TransformedIngestor, ParquetIngestor
        >>> from flamme.transformer.df import ToNumeric
        >>> ingestor = TransformedIngestor(
        ...     ingestor=ParquetIngestor(path="/path/to/df.csv"),
        ...     transformer=ToNumeric(columns=["col1", "col3"]),
        ... )
        >>> ingestor
        TransformedIngestor(
          (ingestor): ParquetIngestor(path=/path/to/df.csv)
          (transformer): ToNumericDataFrameTransformer(columns=('col1', 'col3'))
        )
        >>> df = ingestor.ingest()  # doctest: +SKIP
    """

    def __init__(
        self, ingestor: BaseIngestor | dict, transformer: BaseDataFrameTransformer | dict
    ) -> None:
        self._ingestor = setup_ingestor(ingestor)
        self._transformer = setup_dataframe_transformer(transformer)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping({"ingestor": self._ingestor, "transformer": self._transformer})
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def ingest(self) -> DataFrame:
        df = self._ingestor.ingest()
        return self._transformer.transform(df)



from __future__ import annotations

__all__ = ["ToDatetimeDataFrameTransformer"]

from collections.abc import Sequence

import pandas as pd
from pandas import DataFrame
from tqdm import tqdm

from flamme.transformer.df.base import BaseDataFrameTransformer


class ToDatetimeDataFrameTransformer(BaseDataFrameTransformer):
    r"""Implements a transformer to convert some columns to numeric type.

    Args:
        columns (``Sequence``): Specifies the columns to convert.
        **kwargs: Specifies the keyword arguments for
            ``pandas.to_datetime``.

    Example usage:

    .. code-block:: pycon

        >>> import pandas as pd
        >>> from flamme.transformer.df import ToDatetime
        >>> transformer = ToDatetime(columns=["col1"])
        >>> transformer
        ToDatetimeDataFrameTransformer(columns=('col1',))
        >>> df = pd.DataFrame(
        ...     {
        ...         "col1": ["2020-1-1", "2020-1-2", "2020-1-31", "2020-12-31", "2021-12-31"],
        ...         "col2": [1, 2, 3, 4, 5],
        ...         "col3": ["a", "b", "c", "d", "e"],
        ...     }
        ... )
        >>> df.dtypes
        col1    object
        col2     int64
        col3    object
        dtype: object
        >>> df = transformer.transform(df)
        >>> df.dtypes
        col1    datetime64[ns]
        col2             int64
        col3            object
        dtype: object
    """

    def __init__(self, columns: Sequence[str], **kwargs) -> None:
        self._columns = tuple(columns)
        self._kwargs = kwargs

    def __repr__(self) -> str:
        args = ", ".join([f"{key}={value}" for key, value in self._kwargs.items()])
        if args:
            args = ", " + args
        return f"{self.__class__.__qualname__}(columns={self._columns}{args})"

    def transform(self, df: DataFrame) -> DataFrame:
        for col in tqdm(self._columns, desc="Converting to numeric type"):
            df[col] = pd.to_datetime(df[col], **self._kwargs)
        return df



from __future__ import annotations

__all__ = ["ColumnContinuousAdvancedSection"]

import logging
from collections.abc import Sequence

import numpy as np
from jinja2 import Template
from matplotlib import pyplot as plt
from pandas import Series

from flamme.section.base import BaseSection
from flamme.section.continuous import (
    add_cdf_plot,
    create_boxplot_figure,
    create_histogram_figure,
    create_stats_table,
)
from flamme.section.utils import (
    GO_TO_TOP,
    auto_yscale_continuous,
    compute_statistics,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.figure import figure2html, readable_xticklabels
from flamme.utils.range import find_range

logger = logging.getLogger(__name__)


class ColumnContinuousAdvancedSection(BaseSection):
    r"""Implements a section that analyzes a continuous distribution of
    values.

    Args:
        series: Specifies the series/column to analyze.
        column: Specifies the column name.
        nbins: Specifies the number of bins in the histogram.
        yscale: Specifies the y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'/'symlog'`` scale is chosen based
            on the distribution.
        figsize: Specifies the figure size in inches. The first
            dimension is the width and the second is the height.
    """

    def __init__(
        self,
        series: Series,
        column: str,
        nbins: int | None = None,
        yscale: str = "auto",
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._series = series
        self._column = column
        self._nbins = nbins
        self._yscale = yscale
        self._figsize = figsize

    @property
    def column(self) -> str:
        return self._column

    @property
    def yscale(self) -> str:
        return self._yscale

    @property
    def nbins(self) -> int | None:
        return self._nbins

    @property
    def series(self) -> Series:
        return self._series

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""tuple: The individual figure size in pixels. The first
        dimension is the width and the second is the height."""
        return self._figsize

    def get_statistics(self) -> dict[str, float | int]:
        return compute_statistics(self._series)

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(f"Rendering the continuous distribution of {self._column}")
        stats = self.get_statistics()
        null_values_pct = (
            f"{100 * stats['num_nulls'] / stats['count']:.2f}" if stats["count"] > 0 else "N/A"
        )
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._column,
                "table": create_stats_table(stats=stats, column=self._column),
                "total_values": f"{stats['count']:,}",
                "unique_values": f"{stats['nunique']:,}",
                "null_values": f"{stats['num_nulls']:,}",
                "null_values_pct": null_values_pct,
                "full_histogram": self._create_full_histogram(stats),
                "iqr_histogram": self._create_iqr_histogram(),
                "full_boxplot": self._create_full_boxplot(),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the discrete distribution of values for column <em>{{column}}</em>.

<ul>
  <li> total values: {{total_values}} </li>
  <li> number of unique values: {{unique_values}} </li>
  <li> number of null values: {{null_values}} / {{total_values}} ({{null_values_pct}}%) </li>
</ul>

<p style="margin-top: 1rem;">
<b> Analysis of the distribution </b>

{{full_histogram}}
{{full_boxplot}}
{{table}}
<p style="margin-top: 1rem;">

<p style="margin-top: 1rem;">
<b> Analysis of distribution in the inter-quartile range (IQR) </b>

{{iqr_histogram}}

"""

    def _create_full_boxplot(self) -> str:
        return create_boxplot_figure(
            series=self._series,
            xmin="q0",
            xmax="q1",
            figsize=self._figsize,
        )

    def _create_full_histogram(self, stats: dict) -> str:
        return create_histogram_figure(
            series=self._series,
            column=self._column,
            stats=stats,
            nbins=self._nbins,
            xmin="q0",
            xmax="q1",
            yscale=self._yscale,
            figsize=self._figsize,
        )

    def _create_iqr_histogram(self) -> str:
        return create_histogram_range_figure(
            series=self._series,
            column=self._column,
            nbins=self._nbins,
            xmin="q0.25",
            xmax="q0.75",
            yscale=self._yscale,
            figsize=self._figsize,
        )


def create_histogram_range_figure(
    series: Series,
    column: str,
    nbins: int | None = None,
    yscale: str = "auto",
    xmin: float | str | None = None,
    xmax: float | str | None = None,
    figsize: tuple[float, float] | None = None,
) -> str:
    r"""Creates the HTML code of a histogram figure.

    Args:
        series: Specifies the series/column to analyze.
        column: Specifies the column name.
        nbins: Specifies the number of bins in the histogram.
        yscale: Specifies the y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'`` scale is chosen based on the
            distribution.
        xmin: Specifies the minimum value of the range or its
            associated quantile. ``q0.1`` means the 10% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        xmax: Specifies the maximum value of the range or its
            associated quantile. ``q0.9`` means the 90% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        figsize: Specifies the figure size in inches. The first
            dimension is the width and the second is the height.

    Returns:
        str: The HTML code of the figure.
    """
    array = series.to_numpy()
    if array.size == 0:
        return "<span>&#9888;</span> No figure is generated because the column is empty"
    xmin, xmax = find_range(array, xmin=xmin, xmax=xmax)
    array_range = array[np.logical_and(array >= xmin, array <= xmax)]
    fig, ax = plt.subplots(figsize=figsize)
    ax.hist(array_range, bins=nbins, range=[xmin, xmax], color="tab:blue", alpha=0.9)
    if xmin < xmax:
        ax.set_xlim(xmin, xmax)
    readable_xticklabels(ax, max_num_xticks=100)
    ax.set_title(f"Distribution of values for column {column}")
    ax.set_ylabel("Number of occurrences")
    if yscale == "auto":
        yscale = auto_yscale_continuous(array=array_range, nbins=nbins)
    ax.set_yscale(yscale)
    add_cdf_plot(ax, array=array, nbins=nbins, xmin=xmin, xmax=xmax)
    return figure2html(fig, close_fig=True)



from __future__ import annotations

__all__ = ["ToNumericSeriesTransformer"]


import pandas as pd
from pandas import Series

from flamme.transformer.series.base import BaseSeriesTransformer


class ToNumericSeriesTransformer(BaseSeriesTransformer):
    r"""Implements a ``pandas.Series`` transformer to convert a
    ``pandas.Series`` to numeric type.

    Args:
        **kwargs: Specifies the keyword arguments for
            ``pandas.to_numeric``.

    Example usage:

    .. code-block:: pycon

        >>> import pandas as pd
        >>> from flamme.transformer.series import ToNumeric
        >>> transformer = ToNumeric()
        >>> transformer
        ToNumericSeriesTransformer()
        >>> series = pd.Series(["1", "2", "3", "4", "5"])
        >>> series.dtype
        dtype('O')
        >>> series = transformer.transform(series)
        >>> series.dtype
        dtype('int64')
    """

    def __init__(self, **kwargs) -> None:
        self._kwargs = kwargs

    def __repr__(self) -> str:
        args = ", ".join([f"{key}={value}" for key, value in self._kwargs.items()])
        return f"{self.__class__.__qualname__}({args})"

    def transform(self, series: Series) -> Series:
        return pd.to_numeric(series, **self._kwargs)



from __future__ import annotations

from coola.utils import str_indent


def create_html_report(toc: str, body: str) -> str:
    r"""Creates a HTML report.

    Args:
        toc (str): Specifies the table of contents of the report.
        body (str): Specifies the body of the report

    Returns:
        str: The HTML report.
    """
    return f"""
<!doctype html>
<html>
    <head>
        <link href="https://cdn.jsdelivr.net/npm/bootstrap@5.3.1/dist/css/bootstrap.min.css" rel="stylesheet" integrity="sha384-4bw+/aepP/YC94hEpVNVgiZdgIC5+VKNBQNGCHeKRQN+PtmoHDEXuppvnDJzQIu9" crossorigin="anonymous>"
    </head>
    <body style="margin: 1.5rem;">
    <div id="toc_container">
    <h2>Table of content</h2>
    {str_indent(toc, num_spaces=4)}
    </div>

    \t{str_indent(body, num_spaces=8)}
    </body>
</html>
"""



from __future__ import annotations

__all__ = ["NoRepeatReporter"]

import logging
from pathlib import Path

from coola.utils import str_indent, str_mapping

from flamme.reporter.base import BaseReporter, setup_reporter
from flamme.utils.path import sanitize_path

logger = logging.getLogger(__name__)


class NoRepeatReporter(BaseReporter):
    r"""Implements a reporter that computes the report only once.

    Args:
        reporter (``BaseReporter`` or dict): Specifies the reporter
            or its configuration.
        report_path (``Path`` or str): Specifies the path where to
            save the HTML report.

    Example usage:

    .. code-block:: pycon

        >>> from flamme.analyzer import NullValueAnalyzer
        >>> from flamme.ingestor import ParquetIngestor
        >>> from flamme.transformer.df import SequentialDataFrameTransformer
        >>> from flamme.reporter import Reporter, NoRepeatReporter
        >>> reporter = NoRepeatReporter(
        ...     reporter=Reporter(
        ...         ingestor=ParquetIngestor("/path/to/data.parquet"),
        ...         transformer=SequentialDataFrameTransformer(transformers=[]),
        ...         analyzer=NullValueAnalyzer(),
        ...         report_path="/path/to/report.html",
        ...     ),
        ...     report_path="/path/to/report.html",
        ... )
        >>> report = reporter.compute()  # doctest: +SKIP
    """

    def __init__(
        self,
        reporter: BaseReporter | dict,
        report_path: Path | str,
    ) -> None:
        self._reporter = setup_reporter(reporter)
        self._report_path = sanitize_path(report_path)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "reporter": self._reporter,
                    "report_path": self._report_path,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def compute(self) -> None:
        if self._report_path.is_file():
            logger.warning(
                f"The report ({self._report_path}) already exists and it is not re-computed"
            )
            return
        self._reporter.compute()



from __future__ import annotations

__all__ = ["ParquetIngestor"]

import logging
from pathlib import Path

from pandas import DataFrame, read_parquet

from flamme.ingestor.base import BaseIngestor
from flamme.utils.path import human_file_size, sanitize_path

logger = logging.getLogger(__name__)


class ParquetIngestor(BaseIngestor):
    r"""Implements a parquet DataFrame ingestor.

    Args:
        path (``pathlib.Path`` or str): Specifies the path to the
            parquet file to ingest.
        **kwargs: Additional keyword arguments for
            ``pandas.read_parquet``.

    Example usage:

    .. code-block:: pycon

        >>> from flamme.ingestor import ParquetIngestor
        >>> ingestor = ParquetIngestor(path="/path/to/df.parquet")
        >>> ingestor
        ParquetIngestor(path=/path/to/df.parquet)
        >>> df = ingestor.ingest()  # doctest: +SKIP
    """

    def __init__(self, path: Path | str, **kwargs) -> None:
        self._path = sanitize_path(path)
        self._kwargs = kwargs

    def __repr__(self) -> str:
        args = ", ".join([f"{key}={value}" for key, value in self._kwargs.items()])
        if args:
            args = ", " + args
        return f"{self.__class__.__qualname__}(path={self._path}{args})"

    def ingest(self) -> DataFrame:
        logger.info(
            f"Ingesting parquet data from {self._path} (size={human_file_size(self._path)})..."
        )
        df = read_parquet(path=self._path, **self._kwargs)
        logger.info(f"Data ingested. DataFrame shape: {df.shape}")
        return df



from __future__ import annotations

__all__ = ["nonnan"]

import numpy as np


def nonnan(array: np.ndarray) -> np.ndarray:
    r"""Returns the non-NaN values of an array.

    Args:
        array: Specifies the input array.

    Returns:
        A 1d array with the non-NaN values of the input array.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> from flamme.utils.array import nonnan
    >>> nonnan(np.asarray([1, 2, float("nan"), 5, 6]))
    array([1., 2., 5., 6.])
    >>> nonnan(np.asarray([[1, 2, float("nan")], [4, 5, 6]]))
    array([1., 2., 4., 5., 6.])

    ```
    """
    mask = np.isnan(array)
    return array[~mask]



from __future__ import annotations

__all__ = [
    "df_column_types",
    "find_date_columns_from_dtypes",
    "find_numeric_columns_from_dtypes",
    "read_dtypes_from_schema",
    "series_column_types",
]

import logging
from pathlib import Path

import pyarrow as pa
import pyarrow.parquet as pq
from pandas import DataFrame, Series

from flamme.utils.path import sanitize_path

logger = logging.getLogger(__name__)


def df_column_types(df: DataFrame) -> dict[str, set]:
    r"""Computes the value types per column.

    Args:
        df: Specifies the DataFrame to analyze.

    Returns:
        A dictionary with the value types for each column.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> import pandas as pd
    >>> from flamme.utils.dtype import df_column_types
    >>> df = pd.DataFrame(
    ...     {
    ...         "int": np.array([np.nan, 1, 0, 1]),
    ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
    ...     }
    ... )
    >>> coltypes = df_column_types(df)
    >>> coltypes
    {'int': {<class 'float'>}, 'float': {<class 'float'>}}

    ```
    """
    types = {}
    for col in df:
        types[col] = series_column_types(df[col])
    return types


def series_column_types(series: Series) -> set[type]:
    r"""Computes the value types in a ``pandas.Series``.

    Args:
        series: Specifies the DataFrame to analyze.

    Returns:
        A dictionary with the value types for each column.

    Example usage:

    ```pycon
    >>> import numpy as np
    >>> import pandas as pd
    >>> from flamme.utils.dtype import series_column_types
    >>> coltypes = series_column_types(pd.Series([1.2, 4.2, np.nan, 2.2]))
    >>> coltypes
    {<class 'float'>}

    ```
    """
    return {type(x) for x in series.tolist()}


def read_dtypes_from_schema(path: Path | str) -> dict:
    r"""Read the column data types from the schema.

    Args:
        path: Specifies the path to the schema.

    Returns:
        The mapping of column names and data types.
    """
    path = sanitize_path(path)
    logger.info(f"Reading schema from {path}")
    schema = pq.read_schema(path)
    return {name: dtype for name, dtype in zip(schema.names, schema.types)}


def find_numeric_columns_from_dtypes(dtypes: dict[str, pa.DataType]) -> list[str]:
    r"""Finds the columns with a numeric type.

    Args:
        dtypes: The mapping of column names and data types.

    Returns:
        The columns with a numeric type.
    """
    columns = []
    for col, dtype in dtypes.items():
        if pa.types.is_decimal(dtype) or pa.types.is_floating(dtype) or pa.types.is_integer(dtype):
            columns.append(col)
    return columns


def find_date_columns_from_dtypes(dtypes: dict[str, pa.DataType]) -> list[str]:
    r"""Finds the columns with a date type.

    Args:
        dtypes: The mapping of column names and data types.

    Returns:
        The columns with a date type.
    """
    columns = []
    for col, dtype in dtypes.items():
        if pa.types.is_date(dtype):
            columns.append(col)
    return columns



from __future__ import annotations

__all__ = ["compute_null_per_col"]

import numpy as np
from pandas import DataFrame


def compute_null_per_col(df: DataFrame) -> DataFrame:
    r"""Computes the number and percentage of null values per column.

    Args:
        df (``pandas.DataFrame``): Specifies the DataFrame to analyze.

    Returns:
        ``pandas.DataFrame``: A DataFrame with the number and
            percentage of null values per column.

    Example usage:

    .. code-block:: pycon

        >>> import pandas as pd
        >>> from flamme.utils.null import compute_null_per_col
        >>> df = compute_null_per_col(
        ...     pd.DataFrame(
        ...         {
        ...             "int": np.array([np.nan, 1, 0, 1]),
        ...             "float": np.array([1.2, 4.2, np.nan, 2.2]),
        ...             "str": np.array(["A", "B", None, np.nan]),
        ...         }
        ...     )
        ... )
        >>> df
          column  null  total  null_pct
        0    int     1      4      0.25
        1  float     1      4      0.25
        2    str     2      4      0.50
    """
    null_count = df.isnull().sum().to_frame("count")["count"].to_numpy().astype(int)
    total_count = np.full((df.shape[1],), df.shape[0]).astype(int)
    with np.errstate(invalid="ignore"):
        null_pct = null_count.astype(float) / total_count.astype(float)
    return DataFrame(
        {
            "column": list(df.columns),
            "null": null_count,
            "total": total_count,
            "null_pct": null_pct,
        }
    )



from __future__ import annotations

__all__ = ["StripStringDataFrameTransformer"]

from collections.abc import Sequence

from pandas import DataFrame
from tqdm import tqdm

from flamme.transformer.df.base import BaseDataFrameTransformer


class StripStringDataFrameTransformer(BaseDataFrameTransformer):
    r"""Implements a transformer to strip the strings of some columns.

    Args:
        columns (``Sequence``): Specifies the columns to process.

    Example usage:

    .. code-block:: pycon

        >>> import pandas as pd
        >>> from flamme.transformer.df import StripString
        >>> transformer = StripString(columns=["col1", "col3"])
        >>> transformer
        StripStringDataFrameTransformer(columns=('col1', 'col3'))
        >>> df = pd.DataFrame(
        ...     {
        ...         "col1": [1, 2, 3, 4, 5],
        ...         "col2": ["1", "2", "3", "4", "5"],
        ...         "col3": ["a ", " b", "  c  ", "d", "e"],
        ...         "col4": ["a ", " b", "  c  ", "d", "e"],
        ...     }
        ... )
        >>> df
           col1 col2   col3   col4
        0     1    1     a      a
        1     2    2      b      b
        2     3    3    c      c
        3     4    4      d      d
        4     5    5      e      e
        >>> df = transformer.transform(df)
        >>> df
           col1 col2 col3   col4
        0     1    1    a     a
        1     2    2    b      b
        2     3    3    c    c
        3     4    4    d      d
        4     5    5    e      e
    """

    def __init__(self, columns: Sequence[str]) -> None:
        self._columns = tuple(columns)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(columns={self._columns})"

    def transform(self, df: DataFrame) -> DataFrame:
        for col in tqdm(self._columns, desc="Striping strings"):
            df[col] = df[col].apply(lambda x: x.strip() if isinstance(x, str) else x)
        return df



from __future__ import annotations

__all__ = ["NullColumnDataFrameTransformer"]

import logging

from pandas import DataFrame

from flamme.transformer.df.base import BaseDataFrameTransformer
from flamme.utils.null import compute_null_per_col

logger = logging.getLogger(__name__)


class NullColumnDataFrameTransformer(BaseDataFrameTransformer):
    r"""Implements a ``pandas.DataFrame`` transformer to remove the
    columns that have too many null values.

    Args:
        threshold (float): Specifies the maximum percentage of null
            values to keep columns. If the proportion of null vallues
            is greater or equal to this threshold value, the column
            is removed. If set to ``1.0``, it removes all the columns
            that have only null values.

    Example usage:

    .. code-block:: pycon

        >>> import pandas as pd
        >>> from flamme.transformer.df import NullColumn
        >>> transformer = NullColumn(threshold=1.0)
        >>> transformer
        NullColumnDataFrameTransformer(threshold=1.0)
        >>> df = pd.DataFrame(
        ...     {
        ...         "col1": ["2020-1-1", "2020-1-2", "2020-1-31", "2020-12-31", None],
        ...         "col2": [1, None, 3, None, 5],
        ...         "col3": [None, None, None, None, None],
        ...     }
        ... )
        >>> df = transformer.transform(df)
        >>> df
                 col1  col2
        0    2020-1-1   1.0
        1    2020-1-2   NaN
        2   2020-1-31   3.0
        3  2020-12-31   NaN
        4        None   5.0
    """

    def __init__(self, threshold: float) -> None:
        self._threshold = float(threshold)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(threshold={self._threshold})"

    def transform(self, df: DataFrame) -> DataFrame:
        if df.shape[0] == 0:
            return df
        num_orig_cols = len(df.columns)
        df_null = compute_null_per_col(df)
        columns = df_null[df_null["null_pct"] < self._threshold]["column"].tolist()
        logger.info(
            f"Removing {len(columns) - num_orig_cols:,} columns because they have too "
            f"many null values (threshold={self._threshold})..."
        )
        return df[columns].copy()



from __future__ import annotations

__all__ = ["BaseSection"]

from abc import ABC, abstractmethod
from collections.abc import Sequence


class BaseSection(ABC):
    r"""Defines the base class to manage sections."""

    @abstractmethod
    def get_statistics(self) -> dict:
        r"""Computes the statistics associated to the section.

        Returns:
        -------
            dict: The statistics.
        """

    @abstractmethod
    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        r"""Renders the HTML body associated to the section.

        Args:
        ----
            number (str, optional): Specifies the section number.
                Default: ""
            tags (``Sequence``, optional): Specifies the tags
                associated to the section. Default: ``()``
            depth (int, optional): Specifies the depth in the report.
                Default: ``0``

        Returns:
        -------
            str: The HTML body associated to the section.
        """

    @abstractmethod
    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        r"""Renders the HTML table of content (TOC) associated to the
        section.

        Args:
        ----
            number (str, optional): Specifies the section number
                associated to the section. Default: ""
            tags (``Sequence``, optional): Specifies the tags
                associated to the section. Default: ``()``
            depth (int, optional): Specifies the depth in the report.
                Default: ``0``

        Returns:
        -------
            str: The HTML table of content associated to the section.
        """



from __future__ import annotations

__all__ = ["TableOfContentSection"]

from collections.abc import Sequence

from flamme.section.base import BaseSection


class TableOfContentSection(BaseSection):
    r"""Implement a wrapper section that generates a table of content
    before the section.

    Args:
        section: Specifies the section.
        max_toc_depth: Specifies the maximum level to show in the
            table of content.
    """

    def __init__(self, section: BaseSection, max_toc_depth: int = 1) -> None:
        self._section = section
        self._max_toc_depth = int(max_toc_depth)

    def get_statistics(self) -> dict:
        return self._section.get_statistics()

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        toc = self._section.render_html_toc(
            number=number, tags=tags, depth=0, max_depth=self._max_toc_depth
        )
        body = self._section.render_html_body(number=number, tags=tags, depth=depth)
        return f"{toc}\n\n{body}"

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return self._section.render_html_toc(
            number=number, tags=tags, depth=depth, max_depth=max_depth
        )
